#include <QTimer>

#include <qwt_painter.h>
#include <qwt_plot_canvas.h>
#include <qwt_plot_grid.h>
#include <qwt_plot_curve.h>
#include <qwt_plot_layout.h>
#include <qwt_scale_widget.h>
#include <qwt_scale_draw.h>
#include <qwt_legend.h>

#include "stm32plot.h"

STM32Plot::STM32Plot(QWidget *parent):
    QwtPlot(parent),
    interval(30),
    numPoints(200),
    refreshPeriod(50),
    newXValue(0),
    newYValue(0),
    newZValue(0)
{
    setTitle("Accelerometro");
    setCanvasBackground(Qt::white);

    plotLayout()->setAlignCanvasToScales(true);
    canvas()->setFrameStyle(QFrame::Box|QFrame::Plain);
    canvas()->setLineWidth(1);
    axisWidget(QwtPlot::xBottom)->setMargin(0);
    axisWidget(QwtPlot::yLeft)->setMargin(0);
    axisWidget(QwtPlot::yRight)->setMargin(0);
    axisScaleDraw(QwtPlot::xBottom)->enableComponent(QwtAbstractScaleDraw::Backbone, false);
    axisScaleDraw(QwtPlot::yLeft)->enableComponent(QwtAbstractScaleDraw::Backbone, false);
    axisScaleDraw(QwtPlot::yRight)->enableComponent(QwtAbstractScaleDraw::Backbone, false);

    QPen gridPen;
    gridPen.setStyle(Qt::DashLine);
    gridPen.setColor(Qt::lightGray);
    grid = new QwtPlotGrid();
    grid->setPen(gridPen);
    grid->attach(this);

    QPen xCurvePen;
    xCurvePen.setColor(Qt::blue);
    xCurvePen.setWidth(2);

    QPen yCurvePen;
    yCurvePen.setColor(Qt::red);
    yCurvePen.setWidth(2);

    QPen zCurvePen;
    zCurvePen.setColor(Qt::darkGreen);
    zCurvePen.setWidth(2);

    xCurve = new QwtPlotCurve("X axis");
    xCurve->setPen(xCurvePen);
    xCurve->setRenderHint(QwtPlotCurve::RenderAntialiased);
    xCurve->attach(this);

    yCurve = new QwtPlotCurve("Y axis");
    yCurve->setPen(yCurvePen);
    yCurve->setRenderHint(QwtPlotCurve::RenderAntialiased);
    yCurve->attach(this);

    zCurve = new QwtPlotCurve("Z axis");
    zCurve->setPen(zCurvePen);
    zCurve->setRenderHint(QwtPlotCurve::RenderAntialiased);
    zCurve->attach(this);

    enableAxis(QwtPlot::xBottom, false);

    setAxisTitle(QwtPlot::yLeft, "g force");
    setAxisScale(QwtPlot::yLeft, -2.3, 2.3, 0.5);

    legend = new QwtLegend();

    insertLegend(legend);

    time.resize(numPoints);
    x.resize(numPoints);
    y.resize(numPoints);
    z.resize(numPoints);

    clearData();

    timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(timerTimeout()));
}

STM32Plot::~STM32Plot()
{
    delete grid;
    delete legend;

    delete xCurve;
    delete yCurve;
    delete zCurve;
}

void STM32Plot::clearData()
{
    for (int i = 0; i < numPoints; i++) {
        time[numPoints-1-i] = i;
        x[i] = 0;
        y[i] = 0;
        z[i] = 0;
    }
}

void STM32Plot::timerTimeout()
{
    x.pop_back();
    x.push_front(newXValue);

    y.pop_back();
    y.push_front(newYValue);

    z.pop_back();
    z.push_front(newZValue);

    for(int j = 0; j < numPoints; j++)
        time[j]++;

    setAxisScale(QwtPlot::xBottom, time.last(), time.first());

    xCurve->setSamples(time, x);
    yCurve->setSamples(time, y);
    zCurve->setSamples(time, z);

    replot();
}

void STM32Plot::addValue(double x, double y, double z)
{
    newXValue = x;
    newYValue = y;
    newZValue = z;
}

void STM32Plot::start()
{
    clearData();
    timer->start(refreshPeriod);
}

void STM32Plot::stop()
{
    timer->stop();
}
